@@ -21,8 +21,9 @@ class CronJobTasks:
         now = datetime.datetime.now()
         logger.info(f'Start {func_name} at {now}')
 
-        error = database.test_session()
-        if isinstance(error, str):
+        if not database.get_session():
+            error = f'Error in {func_name}: not database.get_session()...'
+            logger.error(error)
             return error
 
         if database.project:
@@ -32,6 +33,7 @@ class CronJobTasks:
             filter_value = database.CronjobTask.enabled.is_(True)
 
         cronjob_tasks = database.session.query(database.CronjobTask).filter(filter_value).all()
+        database.session.commit()
 
         if not isinstance(cronjob_tasks, list):
             error = f'Error in {func_name}: not isinstance(cronjob_tasks, list)...'
@@ -55,6 +57,7 @@ class CronJobTasks:
                                     database.Task.scheduled_time > start_of_min)
 
             task = database.session.query(database.Task).filter(filter_value).first()
+            database.session.commit()
 
             if task is None:
                 logger.info(f'{func_name}; apply cronjob_task   : {cronjob_task}')
@@ -73,19 +76,24 @@ class CronJobTasks:
                         new_task_dict[task_column] = getattr(cronjob_task, task_column)
                 new_task = database.add_task(new_task_dict)
                 logger.info(f'{func_name}; create new task      : {new_task}')
+
+        database.close_session()
+
         return None
 
-    def start(self):
+    def start(self, cfg=None):
         func_name = f'{self.__class__.__name__}.{sys._getframe().f_code.co_name}()'
 
         raise_exception = False
         try:
 
-            if not isinstance(config.cfg, dict):
-                logger.error(f'Error in {func_name}: not isinstance(config.cfg, dict)...')
-                return
+            if cfg is None:
+                if not isinstance(config.cfg, dict):
+                    logger.error(f'Error in {func_name}: not isinstance(config.cfg, dict)...')
+                    return
+                cfg = config.cfg
 
-            if not database.init():
+            if not database.init(cfg.get('database')):
                 logger.error(f'Error in {func_name}: not database.init()...')
                 return
 
@@ -93,7 +101,7 @@ class CronJobTasks:
             loops_limit = 1
             life_timeout_sec = None
 
-            cronjobs_cfg = config.cfg.get('cronjobs')
+            cronjobs_cfg = cfg.get('cronjobs')
             if isinstance(cronjobs_cfg, dict):
                 raise_exception = cronjobs_cfg.get('raise_exception', raise_exception)
 
@@ -142,5 +150,5 @@ if __name__ == '__main__':
     init_logger()
 
     logger.warning(f'Start {version}')
-    CronJobTasks().start()
+    CronJobTasks().start(config.cfg)
     logger.warning(f'End {version}')
