@@ -4,16 +4,17 @@ import json
 from pg_tasks_queue.Config import cfg as config
 
 from sqlalchemy import create_engine
-# from sqlalchemy import MetaData
 from sqlalchemy import schema as sqlalchemy_schema
 from sqlalchemy import (Boolean, Column, DateTime, ForeignKey, Integer, Text, VARCHAR)
 from sqlalchemy.dialects.postgresql import JSONB
 from sqlalchemy.orm import relationship
 from sqlalchemy.orm import sessionmaker
-from sqlalchemy_utils import database_exists
+# from sqlalchemy_utils import database_exists
 from sqlalchemy.ext.declarative import declarative_base
 from sqlalchemy import and_
 from sqlalchemy import or_
+# from sqlalchemy.ext.declarative import declared_attr
+# from sqlalchemy import MetaData
 
 
 import logging
@@ -21,27 +22,15 @@ logger = logging.getLogger(__name__)
 
 # config.set_config_dir(os.path.join(os.path.abspath(os.path.dirname(__file__)), '..', 'conf'))
 
+database_cfg = None
 db_engine = None
 session = None
 db_schema = None
 project = None
 
-database_cfg = config.cfg.get('database')
-if isinstance(database_cfg, dict):
-    db_schema = database_cfg.get('schema')
-
-Base = declarative_base()
-
-# https://docs.sqlalchemy.org/en/13/orm/extensions/declarative/mixins.html
-# from sqlalchemy.ext.declarative import declared_attr
-# class Base(object):
-#     @declared_attr
-#     def __tablename__(cls):
-#         return cls.__name__.lower()
-#
-#     __table_args__ = {'mysql_engine': 'InnoDB'}
-#
-#     id =  Column(Integer, primary_key=True)
+# database_cfg = config.cfg.get('database')
+# if isinstance(database_cfg, dict):
+#     db_schema = database_cfg.get('schema')
 
 
 def get_repr(_class):
@@ -56,11 +45,30 @@ def get_repr(_class):
     return repr
 
 
-class CronjobTask(Base):
+# # https://docs.sqlalchemy.org/en/13/orm/extensions/declarative/mixins.html
+# class Base(object):
+#     _schema = None
+#     # if db_schema:
+#     #     _schema = db_schema
+#
+#     @declared_attr
+#     def __table_args__(cls):
+#         args = dict()
+#         if cls._schema:
+#             args.update({'schema': cls._schema})
+#         print(f'CronjobTask: args: {args}')
+#         return args
+# Base = declarative_base(cls=Base)
+# Base._schema = db_schema
+# Base = declarative_base(metadata=MetaData(schema=db_schema))
 
+Base = declarative_base()
+
+
+class CronjobTask(Base):
     __tablename__ = 'cronjob'
-    if db_schema:
-        __table_args__ = {'schema': db_schema}
+    # if db_schema:
+    #     __table_args__ = {'schema': db_schema}
 
     task_id = Column(Integer, primary_key=True)
     module = Column(VARCHAR, nullable=False)
@@ -78,10 +86,7 @@ class CronjobTask(Base):
 
 
 class Worker(Base):
-
     __tablename__ = 'worker'
-    if db_schema:
-        __table_args__ = {'schema': db_schema}
 
     worker_id = Column(Integer, primary_key=True)
     project = Column(VARCHAR)
@@ -99,10 +104,7 @@ class Worker(Base):
 
 
 class Task(Base):
-
     __tablename__ = 'task'
-    if db_schema:
-        __table_args__ = {'schema': db_schema}
 
     task_id = Column(Integer, primary_key=True)
     module = Column(VARCHAR, nullable=False)
@@ -132,10 +134,21 @@ class Task(Base):
     def __repr__(self):
         return get_repr(self)
 
+
+class Config(Base):
+
+    __tablename__ = 'config'
+
+    config_id = Column(Integer, primary_key=True)
+    config = Column(JSONB)
+
+    def __repr__(self):
+        return get_repr(self)
+
 # ======================================================================================================================
 
 
-def get_connection_string(database_cfg):
+def get_connection_string():
     host = database_cfg.get('host')
     if host is None:
         logger.error(f"Error in Database.get_connection_string(): host is None...")
@@ -165,13 +178,16 @@ def get_connection_string(database_cfg):
     return connection_string
 
 
-def init(use_project=True, create=True):
+def init(cfg=None, use_project=True, create=True):
 
-    if not isinstance(config.cfg, dict):
-        logger.error(f"Error in Database.init(): not isinstance(config.cfg, dict)...")
-        return False
+    if cfg is None:
+        if not isinstance(config.cfg, dict):
+            logger.error(f"Error in Database.init(): not isinstance(config.cfg, dict)...")
+            return False
+        cfg = config.cfg.get('database')
+    global database_cfg
+    database_cfg = cfg
 
-    database_cfg = config.cfg.get('database')
     if not isinstance(database_cfg, dict):
         logger.error(f'Error in Database.init(): not isinstance(database_cfg, dict)...')
         return False
@@ -189,7 +205,7 @@ def init(use_project=True, create=True):
         logger.error(f'Error in Database.init(): not isinstance(db_schema, str)...')
         return False
 
-    connection_string = get_connection_string(database_cfg)
+    connection_string = get_connection_string()
     if connection_string is None:
         logger.error(f"Error in Database.init(): connection_string is None...")
         return False
@@ -216,22 +232,14 @@ def init(use_project=True, create=True):
             return False
         logger.warning(f"Schema '{db_schema}' created successfully...")
 
-    # https://docs.sqlalchemy.org/en/13/orm/extensions/declarative/api.html
-    # https://docs.sqlalchemy.org/en/13/orm/extensions/declarative/mixins.html
-    # https://docs.sqlalchemy.org/en/13/changelog/migration_11.html#multi-tenancy-schema-translation-for-table-objects
-    # https://docs.sqlalchemy.org/en/13/faq/metadata_schema.html
-
-    # mymetadata = MetaData(schema=db_schema)
-    # global Base
-    # Base = declarative_base(metadata=mymetadata)
+    global Base
+    # Base.metadata.schema = db_schema
+    for table_name in Base.metadata.tables:
+        Base.metadata.tables[table_name].schema = db_schema
 
     if create:
         Base.metadata.create_all(db_engine)
 
-    Session = sessionmaker(bind=db_engine)
-    global session
-    session = Session()
-
     return True
 
 
@@ -283,12 +291,34 @@ def get_html_table(query_list, columns=None, table=None):
     return {'columns': columns, 'values': values, 'types': types}
 
 
-def test_session():
+def get_session(autocommit=False):
+    global session
     if session is None:
-        error = f'Error: Database.session is None...'
-        logger.error(error)
-        return error
-    return None
+        if db_engine is None:
+            logger.error(f'Error in Database.get_session(): db_engine is None...')
+            return False
+        try:
+            Session = sessionmaker(bind=db_engine)
+            session = Session()
+        except Exception as e:
+            logger.exception(f'Exception in Database.get_session(): {type(e)}: {str(e)}...')
+            session = None
+    if session is None:
+        logger.error(f'Error in Database.get_session(): session is None...')
+        return False
+    session.autocommit = autocommit
+    return True
+
+
+def close_session():
+    global session
+    try:
+        if session is not None:
+            session.close()
+    except Exception as e:
+        logger.exception(f'Exception in Database.close_session(): {type(e)}: {str(e)}...')
+    finally:
+        session = None
 
 
 def get_table(_value, _class):
@@ -303,14 +333,18 @@ def add_common_columns(_value):
             _value.project = project
     if 'params' in _value.__table__.columns:
         if _value.params is None:
-            _value.params = '{}'
+            _value.params = dict()
     if 'create_time' in _value.__table__.columns:
         if _value.create_time is None:
             _value.create_time = datetime.datetime.now()
     return _value
 
 
-def add_new_record(_add_value, _class, func=None, commit=True):
+def add_new_record(_add_value, _class, func=None):
+    if session is None:
+        error = f'Error in Database.add_new_record(): session is None...'
+        logger.error(error)
+        return error
     _add_value = get_table(_add_value, _class)
     if not isinstance(_add_value, _class):
         error = f'Error in Database.add_new_record(): not isinstance(_add_value, {_class.__class__.__name__})...'
@@ -319,23 +353,18 @@ def add_new_record(_add_value, _class, func=None, commit=True):
     if func is not None:
         _add_value.module = func.__module__
         _add_value.func = func.__name__
-    if session is None:
-        if not init():
-            error = f'Error in Database.add_new_record(): not init()...'
-            logger.error(error)
-            return error
+
     session.add(add_common_columns(_add_value))
-    if commit:
-        session.commit()
+    session.commit()
     return _add_value
 
 
 def add_cornjob_task(_add_value, func=None, commit=True):
-    return add_new_record(_add_value, CronjobTask, func=func, commit=commit)
+    return add_new_record(_add_value, CronjobTask, func=func)
 
 
-def add_task(_add_value, func=None, commit=True):
-    return add_new_record(_add_value, Task, func=func, commit=commit)
+def add_task(_add_value, func=None):
+    return add_new_record(_add_value, Task, func=func)
 
 
 def get_new_worker():
@@ -352,10 +381,9 @@ def get_new_worker():
 
 def get_new_task(worker=None):
     if session is None:
-        if not init():
-            error = f'Error in Database.add_new_record(): not init()...'
-            logger.error(error)
-            return error
+        error = f'Error in Database.get_new_task(): session is None...'
+        logger.error(error)
+        return error
 
     # 1. status is null and scheduled_time <= NOW()
     filter_value = and_(Task.status.is_(None),
@@ -394,12 +422,60 @@ def get_new_task(worker=None):
             worker.current_task_id = task.task_id
         task.started_time = datetime.datetime.now()
         task.status = 'working'
-        session.commit()
 
+    session.commit()
     return task
 
 
+def get_new_tasks(task_limit, worker=None):
+    if session is None:
+        error = f'Error in Database.get_new_task(): session is None...'
+        logger.error(error)
+        return error
+
+    # 1. status is null and scheduled_time <= NOW()
+    filter_value_1 = and_(Task.status.is_(None),
+                          Task.scheduled_time.isnot(None),
+                          Task.scheduled_time <= datetime.datetime.now())
+
+    # 2. status = 'error' max_retry_count > retry_count
+    filter_value_2 = and_(Task.status == 'error',
+                          or_(Task.retry_count.is_(None), Task.max_retry_count > Task.retry_count))
+
+    # 3. Simple Tasks
+    filter_value_3 = and_(Task.status.is_(None), Task.scheduled_time.is_(None))
+
+    filter_value = or_(filter_value_1, filter_value_2, filter_value_3)
+
+    if project:
+        filter_value = and_(filter_value, Task.project == project)
+
+    query = session.query(Task).with_for_update(skip_locked=True).filter(filter_value).order_by(
+        Task.task_id.asc()).limit(task_limit)
+
+    tasks = query.all()
+    for task in tasks:
+        if worker is None:
+            host_name = socket.gethostname()
+            host_ip = socket.gethostbyname(host_name)
+            task.worker_host = f'{host_ip} : {host_name}'
+        else:
+            task.worker_host = worker.host
+            task.worker_id = worker.worker_id
+            # worker.current_task_id = task.task_id
+        # task.started_time = datetime.datetime.now()
+        task.status = 'working'
+
+    session.commit()
+    return tasks
+
+
 def update_worker_task(_result, task, worker):
+    if session is None:
+        error = f'Error in Database.update_worker_task(): session is None...'
+        logger.error(error)
+        return error
+
     finished_time = datetime.datetime.now()
     error = None
     if isinstance(_result, dict):
@@ -425,25 +501,35 @@ def update_worker_task(_result, task, worker):
             retry_count = 0 if retry_count is None else retry_count
             task.retry_count = retry_count + 1
 
-        failed_tasks = json.loads(worker.failed_tasks)
-        failed_tasks.append(task.task_id)
-        worker.failed_tasks = json.dumps(failed_tasks)
+        if worker:
+            failed_tasks = json.loads(worker.failed_tasks)
+            failed_tasks.append(task.task_id)
+            worker.failed_tasks = json.dumps(failed_tasks)
 
     else:
         task.status = status
         task.result = result
 
-        success_tasks = json.loads(worker.success_tasks)
-        success_tasks.append(task.task_id)
-        worker.success_tasks = json.dumps(success_tasks)
+        if worker:
+            success_tasks = json.loads(worker.success_tasks)
+            success_tasks.append(task.task_id)
+            worker.success_tasks = json.dumps(success_tasks)
+
+    if worker:
+        worker.current_task_id = None
 
-    worker.current_task_id = None
     task.finished_time = finished_time
-
     session.commit()
 
+    return None
+
 
 def set_worker_error(error, worker, task):
+    if session is None:
+        error = f'Error in Database.set_worker_error(): session is None...'
+        logger.error(error)
+        return error
+
     finished_time = datetime.datetime.now()
 
     if task:
@@ -468,7 +554,6 @@ def set_worker_error(error, worker, task):
 
     session.commit()
 
-
 # ======================================================================================================================
 
 
@@ -481,29 +566,37 @@ if __name__ == '__main__':
     init_logger()
     logger.warning(version)
 
-    # get_new_task()
+    cfg = config.cfg.get('database')
 
-    new_cornjob_task = {'schedule': '*/1 * * * *',
-                        'module': 'Modules.TestTasks',
-                        'func': 'raise_exception',  # raise_exception, sleep_false, sleep
+    if not init(cfg):
+        logger.warning(f"not init()")
+    else:
+        if not get_session():
+            logger.warning(f"not get_session()")
+        else:
+            # get_new_task()
+
+            new_cornjob_task = {'schedule': '*/1 * * * *',
+                                'module': 'Modules.TestTasks',
+                                'func': 'raise_exception',  # raise_exception, sleep_false, sleep
+                                'params': {"counter": 4},
+                                'project': 'project_2',
+                                'max_retry_count': 2,
+                                'enabled': True}
+            # new_cornjob_task = CronjobTask(**new_cornjob_task)
+            # from Modules import TestTasks
+            # new_cornjob_task = add_cornjob_task(new_cornjob_task, TestTasks.raise_exception)
+            new_cornjob_task = add_cornjob_task(new_cornjob_task)
+            print(f'new_cornjob_task: {new_cornjob_task}')
+
+            import datetime
+            new_task = {'module': 'Modules.TestTasks',
+                        'func': 'raise_exception',
                         'params': {"counter": 4},
-                        'project': 'project_2',
-                        'max_retry_count': 2,
-                        'enabled': True}
-    # new_cornjob_task = CronjobTask(**new_cornjob_task)
-    # from Modules import TestTasks
-    # new_cornjob_task = add_cornjob_task(new_cornjob_task, TestTasks.raise_exception)
-    new_cornjob_task = add_cornjob_task(new_cornjob_task)
-    print(f'new_cornjob_task: {new_cornjob_task}')
+                        'max_retry_count': 20,
+                        'scheduled_time': datetime.datetime.now(),
+                        'priority': 1}
+            new_task = add_task(new_task)
+            print(f'new_task: {new_task}')
 
-    # import datetime
-    # new_task = {'module': 'Modules.TestTasks',
-    #             'func': 'raise_exception',
-    #             'params': '{"counter": 4}',
-    #             'max_retry_count': 20,
-    #             'scheduled_time': datetime.datetime.now(),
-    #             'priority': 1}
-    # new_task = add_task(new_task)
-    # print(f'new_task: {new_task}')
-
-    # logger.warning(f'init(): {init()}')
+            close_session()
